/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C)  2008 Julian Lyndon-Smith
 *
 * Julian Lyndon-Smith <asterisk@dotr.com>
 *
 * func_shared.c is distributed with no restrictions on usage or
 * redistribution.
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 */

/*! \file
 * 
 * \brief Shared - Manipulate shared variables in a channel datastore
 *
 * \author Julian Lyndon-Smith <asterisk@dotr.com
 *
 * \ingroup functions
 */
 
/*** MODULEINFO
 ***/

#include "asterisk.h"

ASTERISK_FILE_VERSION(__FILE__, "$Revision:$")

#include "asterisk/lock.h"
#include "asterisk/file.h"
#include "asterisk/channel.h"
#include "asterisk/pbx.h"
#include "asterisk/cli.h"
#include "asterisk/module.h"
#include "asterisk/app.h"
#include "asterisk/utils.h"
#include "asterisk/threadstorage.h"
#include "asterisk/global_datastores.h"

struct pbx_shared {
	struct varshead head;
};

static void shared_store_free(void *data)
{
 /* TODO: don't know what to do here */
}

static struct ast_datastore_info shared_store_info = {
	.type = "SHARED",
	.destroy = shared_store_free,
};

static int acf_shared_write(struct ast_channel *chan, const char *cmd, char *data, const char *value)
{
	struct ast_channel *targetchan;
	struct ast_var_t *newvariable;
	struct varshead *headp;

	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(varname);
		AST_APP_ARG(channel);
	);

	if (ast_strlen_zero(data)) {
		ast_log(LOG_WARNING, "SHARED requires at least one argument (Name)\n");
		return -1;
	}

	AST_STANDARD_APP_ARGS(args, data);

	if (!ast_strlen_zero(args.varname)) {
	
		if (!ast_strlen_zero(args.channel)) {
			targetchan = ast_get_channel_by_name_locked(args.channel);

			if (!targetchan) {
				ast_log(LOG_WARNING, "SHARED: Invalid channel name (%s)\n",args.channel);
				return -1;
			}
		} else targetchan = chan; 
		
		struct ast_datastore *ds = ast_channel_datastore_find(targetchan, &shared_store_info, NULL);
		struct pbx_shared *shared = NULL;

		if (!ds) {

			ds = ast_channel_datastore_alloc(&shared_store_info, NULL);

			if (!ds)
				return -1;
	
			shared = ast_calloc(1, sizeof(struct pbx_shared));

			if (!shared) {
				ast_channel_datastore_free(ds);
				return -1;
			}

		ds->data = shared;

			ast_channel_datastore_add(targetchan, ds);

		} else
			shared = ds->data;

		headp = &shared->head;

		AST_LIST_TRAVERSE_SAFE_BEGIN(headp, newvariable, entries) {
			if (!strcasecmp(ast_var_name(newvariable), args.varname)) {
				/* there is already such a variable, delete it */
				AST_LIST_REMOVE_CURRENT(entries);
				ast_var_delete(newvariable);
				break;
			}
		}

		AST_LIST_TRAVERSE_SAFE_END;
			if (value) {
			newvariable = ast_var_assign(args.varname, value);
			AST_LIST_INSERT_HEAD(headp, newvariable, entries);
		}

		if (!ast_strlen_zero(args.channel)) 
				ast_channel_unlock(targetchan);
	}

	return 0;
}

static int acf_shared_read(struct ast_channel *chan, const char *cmd, char *data, char *buf, size_t len)
{
	struct ast_channel *targetchan;
	struct ast_var_t *newvariable;
	struct varshead *headp;

	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(varname);
		AST_APP_ARG(channel);
	);

	*buf = '\0';
	
	if (ast_strlen_zero(data)) {
		ast_log(LOG_WARNING, "SHARED requires at least one argument (Name)\n");
		return -1;
	}

	AST_STANDARD_APP_ARGS(args, data);

	if (!ast_strlen_zero(args.varname)) {

		if (!ast_strlen_zero(args.channel)) {
			targetchan = ast_get_channel_by_name_locked(args.channel);

			if (!targetchan) {
				ast_log(LOG_WARNING, "SHARED: Invalid channel name (%s)\n",args.channel);
				return -1;
			}
		} else targetchan = chan; 
		
		char *s = alloca(strlen(args.varname) + 4);

		if (s) {

			sprintf(s, "${%s}", args.varname);

 struct ast_datastore *ds = ast_channel_datastore_find(targetchan, &shared_store_info, NULL);
        struct pbx_shared *shared = NULL;

        if (!ds) 
			return -1;
            
		shared = ds->data;

        headp = &shared->head;

        AST_LIST_TRAVERSE_SAFE_BEGIN(headp, newvariable, entries) {
            if (!strcasecmp(ast_var_name(newvariable), args.varname)) {
				ast_copy_string(buf,ast_var_value(newvariable),len);
                break;
            }
        }

        AST_LIST_TRAVERSE_SAFE_END;

		}

		if (!ast_strlen_zero(args.channel)) 
			ast_channel_unlock(targetchan);
	}

	return 0;
}

struct ast_custom_function acf_shared = {
	.name = "SHARED",
	.synopsis = "Manipulates shared variables in a channel",
	.syntax = "SHARED(variable[,channel])",
	.desc =
	"  variable  - variable to update\n"
	"  channel - Optional channel to get / set on\n",
	.read = acf_shared_read,
	.write = acf_shared_write,
};

static int unload_module(void)
{
	int res;

	res = ast_custom_function_unregister(&acf_shared);

	return res;
}

static int load_module(void)
{
	int res;

	res = ast_custom_function_register(&acf_shared);

	return res;
}

AST_MODULE_INFO_STANDARD(ASTERISK_GPL_KEY, "Manipulate Shared Variables");


